@@ -7,7 +7,7 @@ use {
     crate::{
         assign::{self, AssignIF},
         cdb::{self, ClauseDBIF},
-        solver::{restart::RestartIF, SolverEvent},
+        solver::{restart, restart::RestartIF, SolverEvent},
         state::{State, StateIF},
         types::*,
     },
@@ -208,9 +208,9 @@ impl Instantiate for Eliminator {
             SolverEvent::NewVar => {
                 let len = self.var_queue.heap.len();
                 self.var.push(LitOccurs::default());
-                self.var_queue.heap.push(len);
-                self.var_queue.idxs.push(len);
-                self.var_queue.idxs[0] = len;
+                self.var_queue.heap.push(len as u32);
+                self.var_queue.idxs.push(len as u32);
+                self.var_queue.idxs[0] = len as u32;
             }
             SolverEvent::Reinitialize => {
                 self.elim_lits.clear();
@@ -230,11 +230,7 @@ impl EliminateIF for Eliminator {
     fn is_running(&self) -> bool {
         self.enable && self.mode == EliminatorMode::Running
     }
-    fn prepare<A, C>(&mut self, asg: &mut A, cdb: &mut C, force: bool)
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-    {
+    fn prepare(&mut self, asg: &mut impl AssignIF, cdb: &mut impl ClauseDBIF, force: bool) {
         debug_assert!(self.enable);
         if self.mode != EliminatorMode::Waiting {
             return;
@@ -244,7 +240,7 @@ impl EliminateIF for Eliminator {
             w.clear();
         }
         for (cid, c) in &mut cdb.iter_mut().enumerate().skip(1) {
-            if c.is_dead() || c.is(Flag::OCCUR_LINKED) {
+            if c.is_dead() || c.is(FlagClause::OCCUR_LINKED) {
                 continue;
             }
             let vec = c.iter().copied().collect::<Vec<_>>();
@@ -253,46 +249,38 @@ impl EliminateIF for Eliminator {
         }
         if force {
             for vi in 1..=asg.derefer(assign::property::Tusize::NumVar) {
-                if asg.var(vi).is(Flag::ELIMINATED) || asg.assign(vi).is_some() {
+                if asg.var(vi).is(FlagVar::ELIMINATED) || asg.assign(vi).is_some() {
                     continue;
                 }
                 self.enqueue_var(asg, vi, true);
             }
         }
     }
-    fn enqueue_var<A>(&mut self, asg: &mut A, vi: VarId, upward: bool)
-    where
-        A: AssignIF,
-    {
+    fn enqueue_var(&mut self, asg: &mut impl AssignIF, vi: VarId, upward: bool) {
         if self.mode != EliminatorMode::Running {
             return;
         }
         let w = &mut self[vi];
-        if !asg.var(vi).is(Flag::ENQUEUED) && w.activity() < self.eliminate_occurrence_limit {
-            asg.var_mut(vi).turn_on(Flag::ENQUEUED);
+        if !asg.var(vi).is(FlagVar::ENQUEUED) && w.activity() < self.eliminate_occurrence_limit {
+            asg.var_mut(vi).turn_on(FlagVar::ENQUEUED);
             self.var_queue.insert(&self.var, vi, upward);
         }
     }
-    fn simplify<A, C, R>(
+    fn simplify(
         &mut self,
-        asg: &mut A,
-        cdb: &mut C,
-        rst: &mut R,
+        asg: &mut impl AssignIF,
+        cdb: &mut impl ClauseDBIF,
+        rst: &mut impl RestartIF,
         state: &mut State,
-    ) -> MaybeInconsistent
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-        R: RestartIF,
-    {
+    ) -> MaybeInconsistent {
         debug_assert_eq!(asg.decision_level(), 0);
         // we can reset all the reasons because decision level is zero.
         #[cfg(feature = "boundary_check")]
         {
-            for v in asg.var_iter().skip(1) {
-                if asg.reason(v.index) != AssignReason::None {
+            for (i, _) in asg.var_iter().enumerate().skip(1) {
+                if asg.reason(i) != AssignReason::None {
                     assert_eq!(
-                        asg.level(v.index),
+                        asg.level(i),
                         asg.derefer(assign::property::Tusize::RootLevel) as DecisionLevel
                     );
                     // asg.reason(v.index) = AssignReason::None;
@@ -300,6 +288,7 @@ impl EliminateIF for Eliminator {
             }
         }
         if self.enable {
+            self.eliminate_grow_limit = rst.derefer(restart::property::Tusize::IntervalScale) / 2;
             self.subsume_literal_limit = (state.config.elm_cls_lim
                 + cdb.derefer(cdb::property::Tf64::DpAverageLBD) as usize)
                 / 2;
@@ -313,15 +302,15 @@ impl EliminateIF for Eliminator {
                 self.stop(asg, cdb);
             }
         } else {
-            asg.propagate(cdb)
-                .map_or(Ok(()), |cc| Err(SolverError::RootLevelConflict(Some(cc))))?;
+            asg.propagate_sandbox(cdb)
+                .map_err(SolverError::RootLevelConflict)?;
         }
         if self.mode != EliminatorMode::Dormant {
             self.stop(asg, cdb);
         }
         cdb.check_size().map(|_| ())
     }
-    fn sorted_iterator(&self) -> Iter<'_, usize> {
+    fn sorted_iterator(&self) -> Iter<'_, u32> {
         self.var_queue.heap[1..].iter()
     }
     fn stats(&self, vi: VarId) -> Option<(usize, usize)> {
@@ -339,11 +328,14 @@ impl EliminateIF for Eliminator {
 
 impl Eliminator {
     /// register a clause id to all corresponding occur lists.
-    pub fn add_cid_occur<A>(&mut self, asg: &mut A, cid: ClauseId, c: &mut Clause, enqueue: bool)
-    where
-        A: AssignIF,
-    {
-        if self.mode != EliminatorMode::Running || c.is(Flag::OCCUR_LINKED) {
+    pub fn add_cid_occur(
+        &mut self,
+        asg: &mut impl AssignIF,
+        cid: ClauseId,
+        c: &mut Clause,
+        enqueue: bool,
+    ) {
+        if self.mode != EliminatorMode::Running || c.is(FlagClause::OCCUR_LINKED) {
             return;
         }
         let evo = self.eliminate_var_occurrence_limit;
@@ -364,7 +356,7 @@ impl Eliminator {
                 w.aborted = true;
                 continue;
             }
-            if !v.is(Flag::ELIMINATED) {
+            if !v.is(FlagVar::ELIMINATED) {
                 if bool::from(*l) {
                     debug_assert!(
                         !w.pos_occurs.contains(&cid),
@@ -389,20 +381,17 @@ impl Eliminator {
                 self.enqueue_var(asg, l.vi(), false);
             }
         }
-        c.turn_on(Flag::OCCUR_LINKED);
+        c.turn_on(FlagClause::OCCUR_LINKED);
         if enqueue {
             self.enqueue_clause(cid, c);
         }
     }
     /// remove a clause id from all corresponding occur lists.
-    pub fn remove_cid_occur<A>(&mut self, asg: &mut A, cid: ClauseId, c: &mut Clause)
-    where
-        A: AssignIF,
-    {
+    pub fn remove_cid_occur(&mut self, asg: &mut impl AssignIF, cid: ClauseId, c: &mut Clause) {
         debug_assert!(self.mode == EliminatorMode::Running);
         debug_assert!(!cid.is_lifted_lit());
         debug_assert!(!c.is_dead());
-        c.turn_off(Flag::OCCUR_LINKED);
+        c.turn_off(FlagClause::OCCUR_LINKED);
         for l in c.iter() {
             if asg.assign(l.vi()).is_none() {
                 self.remove_lit_occur(asg, *l, cid);
@@ -418,17 +407,13 @@ impl Eliminator {
     // Due to a potential bug of killing clauses and difficulty about
     // synchronization between 'garbage_collect' and clearing occur lists,
     // 'stop' should purge all occur lists to purge any dead clauses for now.
-    fn stop<A, C>(&mut self, asg: &mut A, cdb: &mut C)
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-    {
+    fn stop(&mut self, asg: &mut impl AssignIF, cdb: &mut impl ClauseDBIF) {
         let force: bool = true;
         self.clear_clause_queue(cdb);
         self.clear_var_queue(asg);
         if force {
             for c in &mut cdb.iter_mut().skip(1) {
-                c.turn_off(Flag::OCCUR_LINKED);
+                c.turn_off(FlagClause::OCCUR_LINKED);
             }
             for w in &mut self[1..] {
                 w.clear();
@@ -438,16 +423,12 @@ impl Eliminator {
     }
     /// returns false if solver is inconsistent
     /// - calls `clause_queue.pop`
-    pub fn backward_subsumption_check<A, C>(
+    pub fn backward_subsumption_check(
         &mut self,
-        asg: &mut A,
-        cdb: &mut C,
+        asg: &mut impl AssignIF,
+        cdb: &mut impl ClauseDBIF,
         timedout: &mut usize,
-    ) -> MaybeInconsistent
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-    {
+    ) -> MaybeInconsistent {
         debug_assert_eq!(asg.decision_level(), 0);
         while !self.clause_queue.is_empty() || self.bwdsub_assigns < asg.stack_len() {
             // Check top-level assignments by creating a dummy clause
@@ -465,12 +446,12 @@ impl Eliminator {
                 }
                 let best: VarId = if cid.is_lifted_lit() {
                     let vi = Lit::from(cid).vi();
-                    debug_assert!(!asg.var(vi).is(Flag::ELIMINATED));
+                    debug_assert!(!asg.var(vi).is(FlagVar::ELIMINATED));
                     vi
                 } else {
                     let mut tmp = cdb.derefer(cdb::property::Tusize::NumClause);
                     let c = &mut cdb[cid];
-                    c.turn_off(Flag::ENQUEUED);
+                    c.turn_off(FlagClause::ENQUEUED);
                     if c.is_dead() || self.subsume_literal_limit < c.len() {
                         continue;
                     }
@@ -489,14 +470,14 @@ impl Eliminator {
                         } else {
                             w.pos_occurs.len()
                         };
-                        if !v.is(Flag::ELIMINATED) && num_sum < tmp {
+                        if !v.is(FlagVar::ELIMINATED) && num_sum < tmp {
                             b = l.vi();
                             tmp = num_sum;
                         }
                     }
                     b
                 };
-                if best == 0 || asg.var(best).is(Flag::ELIMINATED) {
+                if best == 0 || asg.var(best).is(FlagVar::ELIMINATED) {
                     continue;
                 }
                 self[best].pos_occurs.retain(|cid| !cdb[*cid].is_dead());
@@ -527,8 +508,8 @@ impl Eliminator {
             }
         }
         if asg.remains() {
-            asg.propagate(cdb)
-                .map_or(Ok(()), |cc| Err(SolverError::RootLevelConflict(Some(cc))))?;
+            asg.propagate_sandbox(cdb)
+                .map_err(SolverError::RootLevelConflict)?;
         }
         Ok(())
     }
@@ -537,24 +518,19 @@ impl Eliminator {
     /// # Errors
     ///
     /// if solver becomes inconsistent.
-    fn eliminate<A, C, R>(
+    fn eliminate(
         &mut self,
-        asg: &mut A,
-        cdb: &mut C,
-        rst: &mut R,
+        asg: &mut impl AssignIF,
+        cdb: &mut impl ClauseDBIF,
+        rst: &mut impl RestartIF,
         state: &mut State,
-    ) -> MaybeInconsistent
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-        R: RestartIF,
-    {
+    ) -> MaybeInconsistent {
         let start = state.elapsed().unwrap_or(0.0);
         loop {
             let na = asg.stack_len();
             self.eliminate_main(asg, cdb, rst, state)?;
-            asg.propagate(cdb)
-                .map_or(Ok(()), |cc| Err(SolverError::RootLevelConflict(Some(cc))))?;
+            asg.propagate_sandbox(cdb)
+                .map_err(SolverError::RootLevelConflict)?;
             if na == asg.stack_len()
                 && (!self.is_running()
                     || (0 == self.clause_queue_len() && 0 == self.var_queue_len()))
@@ -571,18 +547,13 @@ impl Eliminator {
         Ok(())
     }
     /// do the elimination task
-    fn eliminate_main<A, C, R>(
+    fn eliminate_main(
         &mut self,
-        asg: &mut A,
-        cdb: &mut C,
-        rst: &mut R,
+        asg: &mut impl AssignIF,
+        cdb: &mut impl ClauseDBIF,
+        rst: &mut impl RestartIF,
         state: &mut State,
-    ) -> MaybeInconsistent
-    where
-        A: AssignIF,
-        C: ClauseDBIF,
-        R: RestartIF,
-    {
+    ) -> MaybeInconsistent {
         debug_assert!(asg.decision_level() == 0);
         if self.mode == EliminatorMode::Dormant {
             return Ok(());
@@ -601,15 +572,15 @@ impl Eliminator {
             }
             while let Some(vi) = self.var_queue.select_var(&self.var, asg) {
                 let v = asg.var_mut(vi);
-                v.turn_off(Flag::ENQUEUED);
-                if !v.is(Flag::ELIMINATED) && asg.assign(vi).is_none() {
+                v.turn_off(FlagVar::ENQUEUED);
+                if !v.is(FlagVar::ELIMINATED) && asg.assign(vi).is_none() {
                     eliminate_var(asg, cdb, self, rst, state, vi, &mut timedout)?;
                 }
             }
             self.backward_subsumption_check(asg, cdb, &mut timedout)?;
             debug_assert!(self.clause_queue.is_empty());
-            asg.propagate(cdb)
-                .map_or(Ok(()), |cc| Err(SolverError::RootLevelConflict(Some(cc))))?;
+            asg.propagate_sandbox(cdb)
+                .map_err(SolverError::RootLevelConflict)?;
             if timedout == 0 {
                 self.clear_clause_queue(cdb);
                 self.clear_var_queue(asg);
@@ -620,10 +591,7 @@ impl Eliminator {
         Ok(())
     }
     /// remove a clause id from literal's occur list.
-    pub fn remove_lit_occur<A>(&mut self, asg: &mut A, l: Lit, cid: ClauseId)
-    where
-        A: AssignIF,
-    {
+    pub fn remove_lit_occur(&mut self, asg: &mut impl AssignIF, l: Lit, cid: ClauseId) {
         let w = &mut self[l.vi()];
         if w.aborted {
             return;
@@ -647,21 +615,18 @@ impl Eliminator {
     /// enqueue a clause into eliminator's clause queue.
     pub fn enqueue_clause(&mut self, cid: ClauseId, c: &mut Clause) {
         if self.mode != EliminatorMode::Running
-            || c.is(Flag::ENQUEUED)
+            || c.is(FlagClause::ENQUEUED)
             || self.subsume_literal_limit < c.len()
         {
             return;
         }
         self.clause_queue.push(cid);
-        c.turn_on(Flag::ENQUEUED);
+        c.turn_on(FlagClause::ENQUEUED);
     }
     /// clear eliminator's clause queue.
-    fn clear_clause_queue<C>(&mut self, cdb: &mut C)
-    where
-        C: ClauseDBIF,
-    {
+    fn clear_clause_queue(&mut self, cdb: &mut impl ClauseDBIF) {
         for cid in &self.clause_queue {
-            cdb[*cid].turn_off(Flag::ENQUEUED);
+            cdb[*cid].turn_off(FlagClause::ENQUEUED);
         }
         self.clause_queue.clear();
     }
@@ -675,10 +640,7 @@ impl Eliminator {
     ///
 
     /// clear eliminator's var queue
-    fn clear_var_queue<A>(&mut self, asg: &mut A)
-    where
-        A: AssignIF,
-    {
+    fn clear_var_queue(&mut self, asg: &mut impl AssignIF) {
         self.var_queue.clear(asg);
     }
     /// return the length of eliminator's var queue.
@@ -688,10 +650,7 @@ impl Eliminator {
 }
 
 #[allow(dead_code)]
-fn check_eliminator<C>(cdb: &C, elim: &Eliminator) -> bool
-where
-    C: ClauseDBIF,
-{
+fn check_eliminator(cdb: &impl ClauseDBIF, elim: &Eliminator) -> bool {
     // clause_queue should be clear.
     // all elements in occur_lists exist.
     // for v in asg {
